/* 
 * $Id: hpscan.c,v 1.2 1998/06/01 20:06:25 wingel Exp $
 * Copyright 1995 by Christer Weinigel (wingel@ctrl-c.liu.se)
 *
 * LEGALESE:
 * I hereby grant anybody the right to do anything with this piece
 * of code. It would be nice if my name is mentioned, but do as you
 * please. And, oh, if this breaks something, it's not my fault,
 * I can't be held responsible for anything you do with your computer.
 *
 * Switches:
 *
 * IO Ports		SW1
 * 0x278 - 0x27f	ON
 * 0x268 - 0x26f	OFF
 *
 * Memory address	SW2	SW3	SW4
 * 0xc0000 - 0xc3fff	ON	ON	ON
 * 0xc4000 - 0xc7fff	OFF	ON	ON
 * 0xc8000 - 0xcbfff	ON	OFF	ON
 * 0xcc000 - 0xcffff	OFF	OFF	ON
 * 0xd0000 - 0xd3fff	ON	ON	OFF
 * 0xd4000 - 0xd7fff	OFF	ON	OFF
 * 0xd8000 - 0xdbfff	ON	OFF	OFF
 * 0xdc000 - 0xdffff	OFF	OFF	OFF
 *
 * IRQ Selection	Jumper next to switches
 * IRQ 5		Bottom of card (towards bus connector)
 * IRQ 7		Top of card
 * None			No jumper
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/miscdevice.h>

#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/io.h>

#include "hpscan.h"

/* 
 * Define this to get debugging compiled in
 */
#undef HPSJ_DEBUG

/*
 * Module parameters
 */
static int io = 0;
MODULE_PARM(io, "i");

#ifdef HPSJ_DEBUG
static int debug = HPSJ_DEBUG;
MODULE_PARM(debug, "i");
#endif

/* 
 * Device major number and name
 */
#define HPSJ_MINOR 140
#define HPSJ_NAME "hpscan"

/*
 * Hardware registers 
 */
#define HPSJ_STATUS (hpsj.iobase+1) /* status register */
#define HPSJ_CTRL (hpsj.iobase+2) /* scanner control/status */
#define HPSJ_SWING (hpsj.iobase+3) /* swing register */
#define HPSJ_COUNT (hpsj.iobase+7) /* shared memory counter */

/*
 * Hardware bits: status register
 */
#define nFAULT		0x08	/* 0 = Error */
#define ONLINE		0x10	/* 1 = On Line (aka SELECT) */
#define ERROR		0x20	/* 1 = Printer out of paper */
#define nACK		0x40	/* 0 = lp ready for next char */
#define nBUSY		0x80	/* 0 = Printer is busy (active low) */

/*
 * Hardware bits: ctrl register
 */
#define STROBE		0x01    /* lp character strobe (1 active) */
#define RDnWR		0x02    /* reading=1, writing=0 */
#define nRESET		0x04    /* reset (0 active) */
#define SLCT		0x08    /* enable data xfer (1 active) */
#define INT_EN		0x10    /* enable interrupts on NACKLDG transition */
#define BUSY_CTL	0x20    /* 1=host ctls busy (data source is scanner) */
				/* 0=scanner ctls busy (data source is host) */
#define	HOST_SRC	0
#define	SCAN_SRC	BUSY_CTL

/* 
 * Hardware bits: swing register 
 */
#define	SWG_EN		0x01	/* Enable swing buffers */
#define	SWG_HND		0x02	/* Swing handshake */
#define	CNT_HL		0x04	/* 1=high, 0=low counter */
#define	CNT_WEN		0x08	/* Enable write to counter */
#define	SWG_SW(x)	((x&0x70)>>4)

#define	SWG_SIZE	16384	/* Size of swing buffer */

/*
 * How long to hog the CPU while waiting
 */
#define HPSJ_HOG (HZ/100)
#if HPSJ_HOG < 2
#undef HPSJ_HOG
#define HPSJ_HOG 2
#endif

/*
 * Information
 */

static struct hpsj
{
    int iobase;
    unsigned char *membase;
#if 1
    unsigned char buffer[SWG_SIZE];
#endif
} hpsj;

static unsigned iobases[] = { 0x278, 0x268, 0 }; /* possible bases */

/*
 * Reset the interface and the scanner 
 */
static void hpsj_reset(void)
{
#ifdef HPSJ_DEBUG
    if (debug > 2)
	printk(KERN_DEBUG HPSJ_NAME ": resetting scanner\n");
#endif

    outb_p(0, HPSJ_SWING);
    outb_p(0, HPSJ_SWING);
  
    outb_p(SCAN_SRC | SLCT, HPSJ_CTRL);
    
    current->state = TASK_INTERRUPTIBLE;
    current->timeout = jiffies + HZ;
    schedule();
  
    outb_p(SCAN_SRC | SLCT | nRESET, HPSJ_CTRL);
}

/* 
 * Set shared memory counter 
 */
static void hpsj_setcount(int count)
{              
    outb_p(CNT_WEN, HPSJ_SWING); /* write to low byte of swing counter */
    outb_p(count, HPSJ_COUNT);
    outb_p(CNT_WEN | CNT_HL, HPSJ_SWING); /* write to high byte */
    outb_p(count >> 8, HPSJ_COUNT);
  
#ifdef HPSJ_DEBUG
    if (debug > 2)
	printk(KERN_DEBUG HPSJ_NAME ": setcount = %d (0x%x)\n", count, count);
#endif
}

/* 
 * Read shared memory counter 
 */
static int hpsj_getcount(void)
{              
    int count;
    
    outb_p(SWG_EN | SWG_HND, HPSJ_SWING);
    count = inb_p(HPSJ_COUNT);
    outb_p(SWG_EN | SWG_HND | CNT_HL, HPSJ_SWING);
    count |= inb_p(HPSJ_COUNT) << 8;
  
#ifdef HPSJ_DEBUG
    if (debug > 2)
	printk(KERN_DEBUG HPSJ_NAME ": getcount = %d (0x%x)\n", count, count);
#endif

  return count;
}

/* 
 * Read data from the scanner, probably should be modified to work in
 * the background, but...
 */
static ssize_t hpsj_read(struct file *file, 
			 char *buf, size_t count, loff_t *ppos)
{
    unsigned long timeout;
    int done_count;
    int r;
    static int used_count, max_count;
    int stop = 20;		/* to avoid infinite loops */

#ifdef HPSJ_DEBUG
    if (debug > 1)
	printk(KERN_DEBUG HPSJ_NAME ": read %d bytes\n", count);
#endif
    
    if (count == 0)
	return 0;
    
    r = verify_area(VERIFY_WRITE, buf, count);  
    if (r)
	return r;

    /* Check if the scanner is in read mode already */
    if ((inb_p(HPSJ_CTRL) & (nRESET | RDnWR)) != (nRESET | RDnWR))
    {                    
	timeout = jiffies + HZ;
	while (jiffies < timeout)
	{
	    if ((inb_p(HPSJ_CTRL) & (RDnWR | STROBE)) != (RDnWR | STROBE))
		break;
      
	    current->state = TASK_INTERRUPTIBLE;
	    current->timeout = jiffies + HPSJ_HOG;
	    schedule();
	}
	if (timeout <= jiffies)
	{
	    printk(KERN_WARNING HPSJ_NAME ": not responding (read strobe)\n");
	    hpsj_reset();
	    return -EIO;
	}

	timeout = jiffies + HZ*3;
	while (jiffies < timeout)
	{
	    if (inb_p(HPSJ_STATUS) & nBUSY)
		break;
	    current->state = TASK_INTERRUPTIBLE;
	    current->timeout = jiffies + HPSJ_HOG;
	    schedule();
	}
	if (timeout <= jiffies)
	{
	    printk(KERN_WARNING HPSJ_NAME ": not responding (read busy)\n");
	    hpsj_reset();
	    return -EIO;
	}
  
	outb_p(SWG_HND, HPSJ_SWING);
	outb_p(0, HPSJ_SWING);
	outb_p(SCAN_SRC | SLCT | nRESET, HPSJ_CTRL);
	outb_p(SCAN_SRC | SLCT | nRESET | RDnWR, HPSJ_CTRL);

	hpsj_setcount(0);
	outb_p(SWG_EN, HPSJ_SWING);

	/* nothing in the buffer */
	used_count = max_count = 0;
    }

    done_count = 0;

    while (done_count < count)
    {
	int n;

	/* bail out if something seems wrong */
	if (stop-- <= 0)
	{
	    printk(KERN_WARNING HPSJ_NAME ": something is wrong, stopping\n");
	    break;
	}

	/* buffer is all used up, and hopefully new data is being transferred */
	if (used_count == max_count)
	{
	    /* No more data to read */
	    if (!(inb_p(HPSJ_SWING) & SWG_EN))
		break;

	    /* wait for the transfer to finish  */
	    timeout = jiffies + HZ*60;
	    while (jiffies < timeout)
	    {
		if (inb_p(HPSJ_SWING) & SWG_HND)
		    break;
	
		if (!(inb_p(HPSJ_CTRL) & BUSY_CTL))
		{
		    outb_p(0, HPSJ_SWING);
		    current->state = TASK_INTERRUPTIBLE;
		    current->timeout = jiffies + HPSJ_HOG;
		    schedule();
		}
		else
		{
		    current->state = TASK_INTERRUPTIBLE;
		    current->timeout = jiffies + HPSJ_HOG;
		    schedule();
		}
	    }
	    if (timeout <= jiffies)
	    {
		printk(KERN_WARNING HPSJ_NAME ": not responding (read data transfer)\n");
		hpsj_reset();
		return -EIO;
	    }

	    used_count = 0;		/* haven't used up anything yet */
	    max_count = hpsj_getcount(); /* how much data did I get */

#ifdef HPSJ_DEBUG
	    if (debug > 2)
		printk(KERN_DEBUG HPSJ_NAME ": read a new buffer, %d bytes\n", max_count);
#endif

	    if (max_count < SWG_SIZE)
		outb_p(2, HPSJ_SWING);
	}

	n = max_count - used_count;
	if (n > count - done_count)
	    n = count - done_count;

#if 1
	memcpy_fromio(hpsj.buffer, hpsj.membase + used_count, n);
	if (copy_to_user(buf + done_count, hpsj.buffer, n))
	    return -EFAULT;
#else
	memcpy_tofs(buf + done_count, hpsj.membase + used_count, n);
#endif
    
	done_count += n;
	used_count += n;

	/* Nothing left in the buffer, but there's more to come. */
	if ((used_count == max_count) && (inb_p(HPSJ_SWING) & SWG_EN))
	    outb_p(SWG_EN, HPSJ_SWING);	/* ask for more data */
    }

    return done_count;
}

/* 
 * Write data to the scanner, probably should be modified to work in
 * the background too, but the scanner application might be confused.
 */
static ssize_t hpsj_write(struct file *file, 
			  const char *buf, size_t count, loff_t *ppos)
{
    unsigned long timeout;
    int done_count;
    int r;

#ifdef HPSJ_DEBUG
    if (debug > 1)
	printk(KERN_DEBUG HPSJ_NAME ": write %d bytes\n", count);
#endif

    if (count == 0)
	return 0;

    r = verify_area(VERIFY_READ, buf, count);  
    if (r)
	return r;

    /* Enter write mode */
    if ((inb_p(HPSJ_CTRL) & (nRESET | RDnWR)) != nRESET)
    {                 
	outb_p(0, HPSJ_SWING);
	outb_p(0, HPSJ_SWING);
	outb_p(SCAN_SRC | SLCT | nRESET, HPSJ_CTRL);
    }

    timeout = jiffies + HZ*10;
    while (jiffies < timeout)
    {
	if (!(inb_p(HPSJ_CTRL) & BUSY_CTL) && (inb_p(HPSJ_STATUS) & nACK))
	    break;

	current->state = TASK_INTERRUPTIBLE;
	current->timeout = jiffies + HPSJ_HOG;
	schedule();
    }
    if (timeout <= jiffies)
    {
	outb_p(0, HPSJ_SWING);
    
	outb_p(HOST_SRC | SLCT | nRESET, HPSJ_CTRL);
	outb_p(HOST_SRC | SLCT | nRESET | RDnWR, HPSJ_CTRL);
	outb_p(SCAN_SRC | SLCT | nRESET, HPSJ_CTRL);

	current->state = TASK_INTERRUPTIBLE;
	current->timeout = jiffies + HZ;
	schedule();

	if (!(inb_p(HPSJ_STATUS) & nACK))
	{
	    printk(KERN_WARNING HPSJ_NAME ": not responding (write acknowledge)\n");
	    hpsj_reset();
	    return -EIO;
	}
    }

    if ((inb_p(HPSJ_SWING) & (SWG_EN | SWG_HND)) != (SWG_EN | SWG_HND))
    {
	outb_p(0, HPSJ_SWING);
    
	timeout = jiffies + HZ*10;
	while (jiffies < timeout)
	{
	    if (inb_p(HPSJ_STATUS) & nBUSY)
		break;
      
	    current->state = TASK_INTERRUPTIBLE;
	    current->timeout = jiffies + HPSJ_HOG;
	    schedule();
	}
	if (timeout <= jiffies)
	{
	    printk(KERN_WARNING HPSJ_NAME ": not responding (write busy)\n");
	    hpsj_reset();
	    return -EIO;
	}
    }

    done_count = 0;

    while (done_count < count)
    {
	int n = count - done_count;

	while (n > SWG_SIZE)
	    n -= SWG_SIZE;

#if 1
	if (copy_from_user(hpsj.buffer, buf + done_count, n))
	    return -EFAULT;
	memcpy_toio(hpsj.membase + SWG_SIZE - n, hpsj.buffer, n);
#else
	memcpy_fromfs(hpsj.membase + SWG_SIZE - n, buf + done_count, n);
#endif

	/* This might take some time... with my original ScanJet,
	 * the maximum timeout neccesary is about 10 seconds,
	 * so just to make sure I've set it to 30 seconds
	 */
	timeout = jiffies + HZ*30;
	while (jiffies < timeout)
	{
	    if (!(inb_p(HPSJ_SWING) & SWG_HND))
		break;
      
	    current->state = TASK_INTERRUPTIBLE;
	    current->timeout = jiffies + HPSJ_HOG;
	    schedule();
	}
	if (timeout <= jiffies)
	{
	    printk(KERN_WARNING HPSJ_NAME ": not responding (write handshake)\n");
	    hpsj_reset();
	    return -EIO;
	}
    
	hpsj_setcount(SWG_SIZE - n);
    
	outb_p(SWG_EN | SWG_HND, HPSJ_SWING);
    
	done_count += n;
    }

    return done_count;
}

/* 
 * A few DOS compatibility ioctls... I don't use the myself
 */
static int hpsj_ioctl(struct inode *inode, struct file *file,
		      unsigned int cmd, unsigned long arg)
{
    int r;
    long value;

    switch (cmd) 
    {
    case HPSJRESET:
	hpsj_reset();
	return 0;
    
    case HPSJERRSTAT:
	if ((r = verify_area(VERIFY_WRITE, (long *)arg, sizeof(long))) != 0)
	    return r;

	value = inb(HPSJ_STATUS);

#if 1
	if (copy_to_user((long *)arg, &value, sizeof(value)))
	    return -EFAULT;
#else
	put_fs_long(value, (long *)arg);
#endif
	
	return 0;
	
    case HPSJBUFSIZE:
	if ((r = verify_area(VERIFY_WRITE, (long *)arg, sizeof(long))) != 0)
	    return r;
	
	value = SWG_SIZE;

#if 1
	if (copy_to_user((long *)arg, &value, sizeof(value)))
	    return -EFAULT;
#else
	put_fs_long(value, (long *)arg);
#endif
	
	return 0;
	
    case HPSJDEBUG:
	return 0;

    default:
	return -EINVAL;
    }
}

/* 
 * Probe for scanner interface card 
 */
__initfunc(static int hpsj_probe(void))
{
    if (check_region(hpsj.iobase, 8)) 
    {
#ifdef HPSJ_DEBUG
	if (debug > 2)
	    printk(KERN_DEBUG HPSJ_NAME ": I/O region at 0x%d - 0x%d busy\n",
		   hpsj.iobase, hpsj.iobase + 8 - 1);
#endif
	return -EIO;
    }

    /* Clear swing register */
    outb_p(0, HPSJ_SWING);
    outb_p(0, HPSJ_SWING);
    
    /* Clear control register */
    outb_p(HOST_SRC | SLCT | nRESET, HPSJ_CTRL);
    outb_p(HOST_SRC | SLCT | nRESET | RDnWR, HPSJ_CTRL);
    outb_p(SCAN_SRC | SLCT | nRESET, HPSJ_CTRL);
    
    /* Write dummy value to swing counter */
    outb_p(CNT_WEN, HPSJ_SWING);
    outb_p(0x55, HPSJ_COUNT);
    outb_p(CNT_WEN | CNT_HL , HPSJ_SWING);
    outb_p(0x2a, HPSJ_COUNT);
	
    /* Try to read it back */
    outb_p(0x02, HPSJ_SWING);
    if (inb_p(HPSJ_COUNT) != 0x55)
	return -EIO;
    outb_p(0x04, HPSJ_SWING);
    if ((inb_p(HPSJ_COUNT) & 0x3f) != 0x2a)
	return -EIO;
	
    /* Find memory base */
    hpsj.membase = (unsigned char *)((SWG_SW(inb_p(HPSJ_SWING)) << 14) | 0xc0000);
	
    printk(KERN_INFO HPSJ_NAME ": interface found at 0x%x, shared memory at 0x%p\n", hpsj.iobase, hpsj.membase);
    
    return 0;
}

/* 
 * Try to open the scanner
 */
static int hpsj_open(struct inode *inode, struct file *file)
{
    if (MOD_IN_USE)
	return -EBUSY;

#if 0
    hpsj_reset();		/* seems like a good thing to do */
#endif

    MOD_INC_USE_COUNT;
  
    return 0;
}

/* 
 * Close the scanner
 */
static int hpsj_close(struct inode * inode, struct file * file)
{
#if 0
    hpsj_reset();		/* clean up afterwards */
#endif

    MOD_DEC_USE_COUNT;

    return 0;
}

/* 
 * Good old file operations, nothing magic here
 */
static struct file_operations hpsj_fops = 
{
    NULL,			/* lseek */
    hpsj_read,
    hpsj_write,
    NULL,			/* readdir */
    NULL,			/* select or poll */
    hpsj_ioctl,
    NULL,			/* mmap */
    hpsj_open,
    hpsj_close,
    NULL,
    NULL,			/* async */
};

#ifdef MODULE

static struct miscdevice hpsj_device = 
{
    HPSJ_MINOR, HPSJ_NAME, &hpsj_fops
};

int init_module(void)
{
    int i;
    int r;
    
    printk(KERN_INFO HPSJ_NAME ": HP ScanJet driver, $Revision: 1.2 $\n");

    if (io)
    {
	hpsj.iobase = io;
	if ((r = hpsj_probe()) < 0)
	{
	    printk(KERN_INFO HPSJ_NAME ": No HP ScanJet interface found at 0x%x, aborting\n", io);
	    return r;
	}
    }
    else
    {
	for (i = 0; iobases[i]; i++)
	{
	    hpsj.iobase = iobases[i];
	    if ((r = hpsj_probe()) >= 0)
		break;
	}
	if (!iobases[i])
	{
#ifdef HPSJ_DEBUG
	    printk(KERN_DEBUG HPSJ_NAME ": No HP ScanJet interface found\n");
#endif
	    return -EIO;
	}
    }

    if ((r = misc_register(&hpsj_device)) != 0)
    {
	printk(KERN_WARNING HPSJ_NAME ": unable to register device, aborting\n");
	return r;
    }

    /* Inform the kernel about the I/O addresses used */
    request_region(hpsj.iobase, 8, HPSJ_NAME);

#ifdef HPSJ_DEBUG
    if (debug)
	printk(KERN_DEBUG HPSJ_NAME ": debug level %d\n", debug);
#endif

    hpsj_reset();

    return 0;
}

void cleanup_module(void)
{
    hpsj_reset();

    release_region(hpsj.iobase, 8);

    misc_deregister(&hpsj_device);
}

#else

#error I haven't tried to compile this driver into the kernel, you're on your own.

#endif

/*
 *
 * $Log: hpscan.c,v $
 * Revision 1.2  1998/06/01 20:06:25  wingel
 * Updated and cleaned up to work with Linux 2.1.103
 *
 * Revision 1.1.1.1  1996/05/16 18:07:15  wingel
 * Imported sources
 *
 * Revision 0.1.1.8  1995/10/09 02:30:12  wingel
 * Minor fix, added an hpsj_reset() to the module_cleanup() function.
 *
 * Revision 0.1.1.7  1995/09/20  12:56:54  wingel
 * Yet another attempt to speed things up.
 *
 * Revision 0.1.1.6  1995/09/20  11:27:17  wingel
 * Fixed warning regarding const pointers for hpsj_write.
 *
 * Revision 0.1.1.5  1995/06/14  12:42:03  wingel
 * Fixed a few timeout problems
 *
 * Revision 0.1.1.4  1995/02/16  17:13:46  wingel
 * Modified to use the new modutils.
 * Moved a few of the comments to other files.
 * Moved the changelog to the end of the file.
 *
 * Revision 0.1.1.3  1995/02/07  16:42:38  wingel
 * Symbolic constants
 *
 * Revision 0.1.1.2  1995/01/28  12:33:24  wingel
 * Cleaned up a few includes.
 *
 * Revision 0.1.1.1  1995/01/28  12:20:22  wingel
 * More comments.
 * Attempts to use RCS (I still don't understand it)
 *
 * Revision 0.1  1995/01/28  12:19:28  wingel
 * First try
 *
 */

